@@ -7,24 +7,8 @@
 from __future__ import print_function
 import json, warnings, os, pathlib
 import numpy as np
-try:
-    from importlib import metadata
-except ImportError:
-    import importlib_metadata as metadata  # python<=3.7
-try:
-    import dcor
-    from sklearn import gaussian_process
-    with open(pathlib.Path(os.path.dirname(__file__)) / '../../versions.py', 'r') as vfile:
-        packages = json.loads(vfile.read())['all']
-        packages = dict(map(lambda s: s.split('>='), packages))
-        if metadata.version('dcor') < packages['dcor']:
-            raise Exception('Version mismatch. Installed version of dcor', metadata.version('dcor'),
-                            'Please install dcor>=', packages['dcor'])
-        if metadata.version('scikit-learn') < packages['scikit-learn']:
-            raise Exception('Version mismatch. Installed version of scikit-learn', metadata.version('scikit-learn'),
-                          'Please install scikit-learn>=', packages['scikit-learn'])
-except Exception as e:
-    warnings.warn(str(e))
+import dcor
+from sklearn import gaussian_process
 from .independence_tests_base import CondIndTest
 
 class GaussProcReg():
@@ -49,7 +33,7 @@ class GaussProcReg():
         Number of null samples to use
 
     cond_ind_test : CondIndTest
-        Conditional independence test that this Gaussian Proccess Regressor will
+        Conditional independence test that this Gaussian Process Regressor will
         calculate the null distribution for.  This is used to grab the
         get_dependence_measure function.
 
@@ -307,7 +291,7 @@ class GaussProcReg():
         Y = [(j, 0)]
         X = [(j, 0)]   # dummy variable here
         Z = parents
-        array, xyz = \
+        array, xyz, _ = \
                 self.cond_ind_test.dataframe.construct_array(
                     X=X, Y=Y, Z=Z,
                     tau_max=tau_max,
@@ -625,7 +609,7 @@ class GPDC(CondIndTest):
             return pval, null_dist
         return pval
 
-    def get_analytic_significance(self, value, T, dim):
+    def get_analytic_significance(self, value, T, dim, xyz):
         """Returns p-value for the distance correlation coefficient.
 
         The null distribution for necessary degrees of freedom (df) is loaded.
@@ -647,6 +631,9 @@ class GPDC(CondIndTest):
         dim : int
             Dimensionality, ie, number of features.
 
+        xyz : array of ints
+            XYZ identifier array of shape (dim,).
+
         Returns
         -------
         pval : float or numpy.nan
